import json
from pydantic import BaseModel, Field
from abc import abstractmethod, ABC


class QueueJobModel(BaseModel):
    """队列任务模型"""

    id: str = Field(description="任务ID")
    data: dict = Field(description="任务数据, json 后数据结构")


class Job:
    """队列Job"""

    def __init__(self, worker, job, reserved):
        self.worker = worker
        self.job = job
        self.reserved = reserved

    def delete(self):
        """释放任务"""
        try:
            self.worker.delete_reserved(self)
        except:
            pass

    def get_reserved_job(self):
        """释放任务标识"""
        return self.reserved

    def to_json(self):
        """获取格式化后的job数据"""

        return json.loads(self.job)

    def get_raw_body(self):
        """获取格式化后的job数据"""

        return self.job


class _ScikitCompact(ABC):

    @abstractmethod
    def push(self, job: QueueJobModel):
        raise NotImplementedError()

    @abstractmethod
    def push_and_response(self, job: QueueJobModel):
        raise NotImplementedError()

    @abstractmethod
    def pop(self, islast: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def update_result(self, task_id: str, data: dict, expired_seconds: int = 1800) -> bool:
        raise NotImplementedError()



import logging
import time
from mltaskfusion.utils import helper
from .vllm import VllmData, VllmModel, TASK_NAME as VLLM_TASK_NAME, VllmTask
from .ollama import OllamaData, OllamaModel, TASK_NAME as OLLAMA_TASK_NAME, OllamaTask
from .stablediffusion import StableDiffusionData, StableDiffusionModel, TASK_NAME as STABLEDIFFUSION_TASK_NAME
from mltaskfusion.db import queue_client, Job

logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s", level=logging.INFO)
task_names = [OLLAMA_TASK_NAME, STABLEDIFFUSION_TASK_NAME, VLLM_TASK_NAME]


class Task:
    def __init__(self, task_name: str):
        self.task_name = task_name

        if self.task_name == OLLAMA_TASK_NAME:
            task = OllamaModel(id=helper.unique_id())
            self.data_class = OllamaData
            self.task_cli = OllamaTask()
        elif self.task_name == STABLEDIFFUSION_TASK_NAME:
            task = StableDiffusionModel(id=helper.unique_id())
            self.data_class = StableDiffusionData
        else:
            task = VllmModel(id=helper.unique_id())
            self.data_class = VllmData
            self.task_cli = VllmTask()

        self.task = task
        self.queue_cli = queue_client(queue_name=task.queue_name)

    def handle(self):
        """handle"""

        logging.info("[task: %s] starting ...", self.task_name)

        while True:
            job = self.queue_cli.pop()

            if not job:
                time.sleep(2)
                continue

            try:
                self._handle_job(job)
            except Exception as ex:
                # TODO 记录日志
                pass

    def _handle_job(self, job: Job):
        """handle job"""

        try:
            job.delete()
            data = self.data_class.model_validate_json(job.get_raw_body())
            logging.info("[task: %s] Processing job: %s ...", self.task_name, data.id)
            result = self.task_cli.handle(data)
            logging.info("[task: %s] Processed job: %s", self.task_name, data.id)
        except Exception as e:
            result = ""
        self.queue_cli.update_result(task_id=data.id, data={"content": result})


def task_pipeline(task_name: str):
    """任务管道

    Parameters
    ----------
    task_name : str
        任务名称
    """

    Task(task_name=task_name).handle()



from typing import Any
from .base import _ScikitCompact, TaskModel
from pydantic import BaseModel, Field

TASK_NAME = "stablediffusion"
QUEUE_NAME = "stablediffusion-vb5o"


class StableDiffusionModel(TaskModel):
    """stablediffusion"""

    name: str = TASK_NAME
    queue_name: str = QUEUE_NAME


class StableDiffusionData(BaseModel):
    prompt: str = Field(max_length=4096)
    images: list = Field(default=[], description="list of image, format: image url, bytes, or PIL Image")


class StableDiffusionTask(_ScikitCompact):
    """StableDiffusion task"""

    def handle(self, data: StableDiffusionData) -> Any:
        pass



import io
import base64
import os
import numpy as np
from PIL import Image, PngImagePlugin
from PIL.Image import Image as PILImage
from typing import Any
from urllib.request import urlopen


def load_image(data: Any) -> Image:
    """加载图片内容

    Parameters
    ----------
    data : Any

    Returns
    -------
    Image

    Raises
    ------
    ValueError
    """

    if isinstance(data, PILImage):
        img = data
    elif isinstance(data, bytes):
        img = Image.open(io.BytesIO(data))
    elif isinstance(data, np.ndarray):
        img = Image.fromarray(data)
    elif os.path.isfile(data):
        try:
            img = Image.open(data)
        except:
            raise ValueError("Input type {} is not supported.".format(type(data)))
    else:
        try:
            img = Image.open(urlopen(data))
        except:
            raise ValueError("Input type {} is not supported.".format(type(data)))

    try:
        img = img.convert("RGB")
    except:
        pass

    return img


def b64_img(image: PILImage) -> str:
    return "data:image/png;base64," + raw_b64_img(image)


def raw_b64_img(image: PILImage) -> str:
    # XXX controlnet only accepts RAW base64 without headers
    with io.BytesIO() as output_bytes:
        metadata = None
        for key, value in image.info.items():
            if isinstance(key, str) and isinstance(value, str):
                if metadata is None:
                    metadata = PngImagePlugin.PngInfo()
                metadata.add_text(key, value)
        image.save(output_bytes, format="PNG", pnginfo=metadata)

        bytes_data = output_bytes.getvalue()

    return str(base64.b64encode(bytes_data), "utf-8")


def concat(im1: PILImage, im2: PILImage) -> PILImage:
    im1 = im1.convert("RGBA")
    im2 = im2.convert("RGBA")
    new_image = Image.new(mode="RGBA", size=(max(im1.width, im2.width), im1.height + im2.height))
    new_image.paste(im1, (0, 0))
    new_image.paste(im2, (0, im1.height))

    return new_image



from abc import ABC, abstractmethod
from typing import Any
from pydantic import BaseModel, Field


class TaskModel(BaseModel):
    """任务基础模型"""

    name: str = Field(description="任务名称")
    queue_name: str = Field(description="队列名称")
    status: str = Field(default="pending", description="任务状态", enum=["pending", "running", "success", "failed"])
    data: dict = Field(default={}, description="任务数据", example={"a": 1, "b": 2}, title="任务数据")


class BaseData(BaseModel):
    """基础 数据结构"""

    id: str = Field(description="任务ID")


class _ScikitCompact(ABC):

    @abstractmethod
    def handle(self, data: Any):  # -> Any:
        raise NotImplementedError()



import json
import time
import redis
from typing import Any, Union
from mltaskfusion.utils import config, helper
from .base import _ScikitCompact, Job, QueueJobModel


class PyRedis(redis.Redis):
    """redis client"""

    def __init__(self):
        host = config.config("REDIS_HOST", "127.0.0.1")
        pwd = config.config("REDIS_PASSWORD", None)
        port = config.config("REDIS_PORT", 6379)

        redis.Redis.__init__(self, host=host, port=int(port), password=pwd)


class RedisQueue(PyRedis, _ScikitCompact):
    """队列"""

    def __init__(self, queue_name="ml-default", retry_after=0):
        """

        Attributes
        ------------
        queue_name
            str

        retry_after
            int 多少秒后重新进入队列,为0时不进入重试队列.
        """
        super(RedisQueue, self).__init__()
        self.queue_name = queue_name
        self.tries = 3
        self.retry_after = retry_after

    def call_lua_script(self, lua_script: str, keys: Union[list, None] = None, args: Union[list, None] = None) -> Any:
        """调用lua脚本

        Parameters
        ----------
        lua_script : str
            脚本
        keys : list, optional
            keys参数集合, by default None
        args : list, optional
            args参数集合, by default None

        Returns
        -------
        Any
        """

        script = self.register_script(lua_script)
        return script(keys=keys if keys else [], args=args if args else [])

    def size(self):
        """获取队列任务数量"""
        lua_script = """
            return redis.call('llen', KEYS[1]) + redis.call('zcard', KEYS[2]) + redis.call('zcard', KEYS[3])
        """
        return self.call_lua_script(
            lua_script=lua_script, keys=[self.queue_name, self.queue_name + ":delayed", self.queue_name + ":reserved"]
        )

    def push(self, job: QueueJobModel) -> int:
        """添加任务

        Parameters
        ------------
        job
            dict

        Returns
        ------------
        int
        """

        job = job.data
        if int(job.get("max_tries", self.tries)) <= int(job.get("attempts", 0)):
            return 0

        lua_script = """
            -- Push the job onto the queue...
            redis.call('rpush', KEYS[1], ARGV[1])
            -- Push a notification onto the "notify" queue...
            redis.call('rpush', KEYS[2], 1)
            return 1
        """

        return self.call_lua_script(
            lua_script=lua_script, keys=[self.queue_name, self.queue_name + ":notify"], args=[self._payload(job)]
        )

    def pop(self, islast: bool = False) -> Job:
        """获取队列任务

        Parameters
        -----------
        islast : bool
            是否删除并返回存储在 中的 key 列表的最后一个元素

        Returns
        -----------
        Job|None
        """

        lua_script = """
            -- Pop the first job off of the queue...
            local job = redis.call('%s', KEYS[1])
            local reserved = false

            if(job ~= false) then
                -- Increment the attempt count and place job on the reserved queue...
                reserved = cjson.decode(job)
                if (reserved['max_tries'] > reserved['attempts']) then
                    reserved['attempts'] = reserved['attempts'] + 1
                    reserved = cjson.encode(reserved)
                    redis.call('zadd', KEYS[2], ARGV[1], reserved)
                end
                redis.call('%s', KEYS[3])
            end

            return {job, reserved}
        """
        command = "rpop" if islast else "lpop"
        lua_script = lua_script % (command, command)
        self.migrate()
        result = self.call_lua_script(
            lua_script=lua_script,
            keys=[self.queue_name, self.queue_name + ":reserved", self.queue_name + ":notify"],
            args=[self._availableat(self.retry_after)],
        )

        if result[0] is not None:
            return Job(worker=self, job=result[0], reserved=result[1])

        return None

    def migrate(self):
        """迁移重试任务"""
        self._migrate_expired_jobs(self.queue_name + ":delayed", self.queue_name)

        if self.retry_after > 0:
            self._migrate_expired_jobs(self.queue_name + ":reserved", self.queue_name)

    def _migrate_expired_jobs(self, from_key: str, to_key: str):
        """迁移等待重试任务

        Parameters
        -------------
        from_key
            str

        to_key
            str

        Returns
        -------------
        mixed
        """

        lua_script = """
            -- Get all of the jobs with an expired "score"...
            local val = redis.call('zrangebyscore', KEYS[1], '-inf', ARGV[1])

            -- If we have values in the array, we will remove them from the first queue
            -- and add them onto the destination queue in chunks of 100, which moves
            -- all of the appropriate jobs onto the destination queue very safely.
            if(next(val) ~= nil) then
                redis.call('zremrangebyrank', KEYS[1], 0, #val - 1)

                for i = 1, #val, 100 do
                    redis.call('rpush', KEYS[2], unpack(val, i, math.min(i+99, #val)))
                    -- Push a notification for every job that was migrated...
                    for j = i, math.min(i+99, #val) do
                        redis.call('rpush', KEYS[3], 1)
                    end
                end
            end

            return val
        """

        result = self.call_lua_script(
            lua_script=lua_script, keys=[from_key, to_key, to_key + ":notify"], args=[self._availableat()]
        )
        return result

    def delete_reserved(self, job):
        """释放重试任务

        Parameters
        ----------------
        job
            Job

        Returns
        ----------------
        int
        """

        return self.zrem(self.queue_name + ":reserved", job.get_reserved_job())

    def delete_all_reserved(self):
        """释放所有重试任务

        Returns
        ----------------
        int
        """

        return self.delete(self.queue_name + ":reserved")

    def delete_and_release(self, job, delay=0):
        lua_script = """
            -- Remove the job from the current queue...
            redis.call('zrem', KEYS[2], ARGV[1])

            -- Add the job onto the "delayed" queue...
            redis.call('zadd', KEYS[1], ARGV[2], ARGV[1])

            return true
        """

        return self.call_lua_script(
            lua_script=lua_script,
            keys=[self.queue_name + ":delayed", self.queue_name + ":reserved"],
            args=[job.get_reserved_job(), self._availableat(delay)],
        )

    def _payload(self, job: dict):
        """组装job

        Parameters
        -------------
        job
            dict

        Returns
        -------------
        str
        """

        if job.get("id", None) is None:
            job["id"] = helper.unique_id()

        job["max_tries"] = self.tries
        job["attempts"] = job.get("attempts", 0)

        return json.dumps(job)

    def _availableat(self, delay: int = 0):
        return int(time.time()) + delay

    def get_cache_key(self, task_id: str) -> str:
        """获取任务存储key

        Parameters
        ----------
        task_id : str
            任务ID

        Returns
        -------
        str
        """
        return self.queue_name + ":" + task_id

    def push_and_response(self, job: QueueJobModel, seconds: int = 30) -> Any:
        """推送队列并获取查询结果

        Parameters
        ----------
        task_id : MlNode
            任务ID
        seconds : int, optional
            等待时间, by default 30
        """
        self.push(job=job)
        expired_time = int(time.time()) + seconds
        lua_script = """
            -- get result...
            local result = redis.call('get', KEYS[1])

            if(result ~= nil) then
                redis.call('del', KEYS[1])
            end

            return result
        """
        result = None

        while expired_time > time.time():
            response = self.call_lua_script(lua_script=lua_script, keys=[self.get_cache_key(job.id)])

            if response is not None:
                result = json.loads(response)
                break

            time.sleep(0.1)
            continue

        return result

    def update_result(self, task_id: str, data: dict, expired_seconds: int = 1800) -> bool:
        """更新任务结果

        Parameters
        ----------
        task_id : str
            任务标识
        data : dict
            结果
        expired_seconds : int
            过期时间

        Returns
        -------
        bool
        """

        return self.setex(self.get_cache_key(task_id=task_id), time=expired_seconds, value=json.dumps(data))



from .redis_db import RedisQueue
from .base import _ScikitCompact, QueueJobModel, Job


def queue_client(driver: str = "redis", queue_name="ml-default") -> _ScikitCompact:
    """队列客户端

    Parameters
    ----------
    driver : str, optional
        - `"redis"`: `redis 驱动`

    queue_name : str, optional
        队列名称, by default "ml-default"

    Returns
    -------
    _type_
        _description_
    """

    if driver == "redis":
        pass

    return RedisQueue(queue_name=queue_name)


__all__ = ["QueueJobModel", "Job", "queue_client"]



import os
import argparse
import threading
from mltaskfusion.task import task_pipeline, task_names


def main():
    parser = argparse.ArgumentParser(description="Command line tool for model selection.")
    parser.add_argument("--config", type=str, required=True, help="配置文件")
    parser.add_argument("--models", type=str, required=True, help="模型名称，多个使用英文,分隔.")

    args = parser.parse_args()
    models = [model for model in args.models.split(",") if model in task_names]

    os.environ["CONFIG_FILE"] = args.config
    threads = []

    for model in models:
        thread = threading.Thread(target=task_pipeline, args=(model,))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()


if __name__ == "__main__":
    main()



from pydantic import Field
from typing import Any
from openai import OpenAI
from mltaskfusion.utils import config, image

from .base import _ScikitCompact, BaseData, TaskModel

TASK_NAME = "vllm"
QUEUE_NAME = "vllm-j76g"


class VllmModel(TaskModel):
    """vllm"""

    name: str = TASK_NAME
    queue_name: str = QUEUE_NAME


class VllmData(BaseData):
    """vllm data"""

    prompt: str = Field(max_length=4096)
    images: list = Field(default=[], description="list of image urls")
    max_tokens: int = 4096
    ml_model_name: str = "openbmb/MiniCPM-Llama3-V-2_5"


class VllmTask(_ScikitCompact):
    """vllm task"""

    def __init__(self):
        self.client = OpenAI(
            api_key="EMPTY",
            base_url=config.config("VLLM_BASE_URL", "http://127.0.0.1:8000/v1"),
        )

    def handle(self, data: VllmData) -> Any:
        new_image = None

        for img in data.images:
            if not new_image:
                new_image = image.load_image(img)
            else:
                new_image = image.concat(im1=new_image, im2=image.load_image(img))

        content = []

        if new_image:
            content.append(
                {
                    "type": "image_url",
                    "image_url": {"url": image.b64_img(image=new_image), "detail": "low"},
                }
            )

        content.append(
            {
                "type": "text",
                "text": data.prompt,
            }
        )

        chat_completion_from_base64 = self.client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": content,
                }
            ],
            model=data.ml_model_name,
            max_tokens=data.max_tokens,
        )
        content = chat_completion_from_base64.choices[0].message.content.split("<|eot_id|>", maxsplit=1)[0]
        return content



#!/usr/bin/env python

import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

with open("requirements.txt", encoding="utf-8") as f:
    requireds = f.read().splitlines()

setup(
    name="mltaskfusion",
    version="0.1.6",
    long_description=open("README.md", encoding="utf-8").read().strip(),
    long_description_content_type="text/markdown",
    packages=find_packages(include=["mltaskfusion", "mltaskfusion.*"]),
    install_requires=requireds,
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    entry_points={"console_scripts": ["mltaskfusion-cli=mltaskfusion.cmd.cli:main"]},
)



"""系统配置"""

import os
from typing import OrderedDict
import dotenv

_CONFIG_DATA = OrderedDict()


def config(key, default=""):
    """获取.env配置内容

    Parameters
    ------------
    key
        str 配置key

    default
        any 默认值

    Returns
    ------------
    any
    """

    global _CONFIG_DATA

    if not _CONFIG_DATA:
        _CONFIG_DATA = dotenv.dotenv_values(dotenv.find_dotenv(os.environ.get("CONFIG_FILE")))

    return _CONFIG_DATA.get(key, default)
